function varargout = makedop853(vfieldname,event,varargin)
%MAKEDOP853 Compile and link the MEX file for the ODE.
%   [MEXFILE,MEXPATH] = MAKEDOP853(VFNAME,EVENT,SRCPATH,EXEPATH)
%       VFNAME (req) : A string that contains the name of the CPP file with
%                      the vector field. DO NOT include the file extension.
%                      The file extension is assumed to be .cpp
%       EVENT (req)  : Boolean 0 or 1 to indicate whether an event function
%                      is provided by the user or not
%       SRCPATH (opt): Path for the dop853 source files. This is an
%                      optional input. The default is assumed to be PWD.
%       EXEPATH (opt): Path where the final MEX file should be saved. This
%                      is also an optional input and defaults to PWD.
%
%       MEXFILE      : The name of the created MEX file.
%       MEXPATH      : Location of the created MEX file.
%
%   MAKEDOP853 without any input arguments prints this help file.

%   Madhusudhan Venkadesan, 11-March-2008

VERBOSE = 0;
try
    if ispc
        objext = 'obj';
    elseif isunix
        objext = 'o';
    end

    cppext = 'cpp';

    templatefileslist = {'EventObj',...
        'EventPoint',...
        'OptionsObj',...
        'globals',...
        'integrator',...
        'machar',...
        'mexdop853',...
        'templates',...
        'utilfuncs'};

    switch nargin
        case {0,1}
            makedop853_doc;
            varargout = {};
            return
        case 2
            [userdir,vfieldname] = fileparts(vfieldname);
            if isempty(userdir)
                userdir = pwd;
            end
            srcdir = pwd;
        case 3
            srcdir = varargin{1};
            if ~isdir(srcdir)
                error('MAKEDOP853:PathNotFound','No such path for source files.');
            end
            userdir = pwd;
        case 4
            srcdir = varargin{1};
            userdir = varargin{2};
            if ~isdir(srcdir)
                error('MAKEDOP853:PathNotFound','No such path for source files.');
            end
            if ~isdir(userdir)
                error('MAKEDOP853:PathNotFound','No such path for final MEX file.');
            end
        otherwise
            error('MAKEDOP853:IncorrectInputs','Too many inputs provided.');
    end
    if (event>0)
        fileslist = templatefileslist;
    elseif (event==0)
        fileslist = [templatefileslist,{'eventfunc'}];
    else
        error('MAKEDOP853:UnknownEvents','The second input (EVENT) should be either 0 or 1');
    end


    if (nargout>2)
        error('MAKEDOP853:IncorrectOutputs','Too many outputs requested.');
    end

    objdir = fullfile(srcdir,'compiledobjects','');
    if ~isdir(objdir)
        if VERBOSE
            disp('Creating folder to store compiled object files...');
        end
        mkdir(objdir);
    end

    outputname = ['dop853_' vfieldname];

    vffdate = dir(fullfile(userdir,[vfieldname '.' cppext]));
    mexdate = dir(fullfile(userdir,[outputname '.' mexext]));

    if isempty(vffdate)
        error('MAKEDOP853:MissingVfieldfunc',['There is no file named ' vfieldname '.' cppext]);
    end

    if ~isempty(mexdate) && (datenum(mexdate.date) > datenum(vffdate.date))
        if VERBOSE
            disp([vfieldname '.' cppext ' has not changed since last build of ' outputname '.' mexext]);
        end
        mexfilename =  fullfile(userdir,[outputname '.' mexext]);
        [mexpath,mexfile] = fileparts(mexfilename);
        varargout(1) = {mexfile};
        if nargout==2
            varargout(2) = {mexpath};
        end
        return
    end

    warncount = 0;
    objstring = [];

    for i = 1:length(fileslist)
        curfile = deblank(fileslist{i}); % deblank just removes any blanks in the file name
        if ~(exist(fullfile(srcdir,[curfile '.' cppext]),'file')==2)
            warning('MAKEDOP853:MissingCPPFile',['Missing file ' curfile '.' cppext '. Not compiling it']);
            warncount = warncount+1;
            continue;
        end
        cppdate = dir(fullfile(srcdir,[curfile '.' cppext]));
        objdate = dir(fullfile(objdir,[curfile '.' objext]));
        if ~isempty(objdate) && (datenum(objdate.date) > datenum(cppdate.date)) && VERBOSE
            disp(['No changes in ' curfile '.' cppext ' since last compile. Skipping it...']);
            continue;
        end
        disp(['Compiling ' curfile  '.' cppext ' ...'])
        feval(@mex,'-c',fullfile(srcdir,[curfile '.' cppext]),'-outdir',objdir);
    end

    % There will be a bunch of strange looking '''' in several places below.
    % That is because, the path names (especially on Windows and Mac OS) can
    % have a space. Blank spaces of that nature will confuse the eval commands.
    % A cleverer alternative will be to use feval. However, we will need to
    % know the number of arguments we want to pass apriori. That is painful in
    % case the DOP853 code is itself in flux. We want to be able to specify the
    % files to be compiled in fileslist and be done. That's why this somewhat
    % convoluted way of calling mex.
    for i = 1:length(fileslist)
        curfile = deblank(fileslist{i});
        objstring = [objstring, ' ', '''', fullfile(objdir, [curfile  '.' objext]), ''''];
    end
    vfieldcppfile = fullfile(userdir,[vfieldname '.' cppext]);
    mexfiles = [objstring ' ' '''' vfieldcppfile ''''];

    if warncount>0
        disp(' ');
        disp('Some source files missing, so the linking');
        disp('process is probably going to fail.');
        disp('Trying it anyway...');
        disp(' ');
    end

    % Before MEX-ing, let's make sure that the output file is not sitting
    % in memory currently.
    clear(outputname);
    
    if (VERBOSE)
        eval(['mex -v -output ' '''' fullfile(userdir,outputname) '''' mexfiles]);
    else
        eval(['mex -output ' '''' fullfile(userdir,outputname) '''' mexfiles]);
    end

    mexfilename =  fullfile(userdir,[outputname '.' mexext]);
    [mexpath,mexfile] = fileparts(mexfilename);
    varargout(1) = {mexfile};
    if nargout==2
        varargout(2) = {mexpath};
    end

catch
    rethrow(lasterror);
end

end

function makedop853_doc()

fprintf('   MAKEDOP853 Compile and link the MEX file for the ODE.\n');
fprintf('      [MEXFILE,MEXPATH] = MAKEDOP853(VFNAME,EVENT,SRCPATH,EXEPATH)\n');
fprintf('          VFNAME (req) : A string that contains the name of the CPP file with\n');
fprintf('                         the vector field. Do not include the file extension.\n');
fprintf('                         The file extension is assumed to be .cpp\n');
fprintf('          EVENT (req)  : Boolean 0 or 1 to indicate whether an event function\n');
fprintf('                         is provided by the user or not\n');
fprintf('          SRCPATH (opt): Path for the dop853 source files. This is an\n');
fprintf('                         optional input. The default is assumed to be PWD.\n');
fprintf('          EXEPATH (opt): Path where the final MEX file should be saved. This\n');
fprintf('                         is also an optional input and defaults to PWD.\n');
fprintf('   \n');
fprintf('          MEXFILE      : The name of the created MEX file.\n');
fprintf('          MEXPATH      : Location of the created MEX file.\n');
fprintf('   \n');
fprintf('      MAKEDOP853 without any input arguments prints this help file.\n');
fprintf('   \n');
fprintf('      Madhusudhan Venkadesan, 11-March-2008\n');

end

